#ifndef linearSystemSpectral_H
#define linearSystemSpectral_H



#include <Eigen/Core>
#include <Eigen/SparseCore>
#include <Eigen/SparseCholesky>


#include <MatOp/SparseGenMatProd.h>
#include <MatOp/SparseSymShiftSolve.h>
#include <GenEigsSolver.h>
#include <SymEigsShiftSolver.h>
#include <SymGEigsSolver.h>
#include <MatOp/SparseCholesky.h>

using namespace Spectra;

#include <vector>
#include <map>



template< class float_t >
class linearSystemSpectral {
    std::vector< std::map< unsigned int , float_t > > _ASparse;

    Eigen::SparseMatrix< float_t > _A , _At , _AtA;

    unsigned int _rows , _columns;

public:
    linearSystemSpectral() {
        _rows = _columns = 0;
    }
    linearSystemSpectral( int rows , int columns ) {
        setDimensions(rows , columns);
    }
    ~linearSystemSpectral() {
    }

    void setDimensions( int rows , int columns ) {
        _rows = rows; _columns = columns;
        _ASparse.clear();
        _ASparse.resize(_rows);
    }

    float_t & A(unsigned int row , unsigned int column) {
        return _ASparse[row][column];
    }

    void preprocess() {
        // convert ad-hoc matrix to Eigen sparse format:
        {
            _A.resize(_rows , _columns);
            std::vector< Eigen::Triplet< float_t > > triplets;
            for( unsigned int r = 0 ; r < _rows ; ++r ) {
                for( typename std::map< unsigned int , float_t >::const_iterator it = _ASparse[r].begin() ; it != _ASparse[r].end() ; ++it ) {
                    unsigned int c = it->first;
                    double val = it->second;
                    triplets.push_back( Eigen::Triplet< float_t >(r,c,val) );
                }
            }
            _A.setFromTriplets( triplets.begin() , triplets.end() );
        }
        _At = _A.transpose();
        _AtA = _At * _A;
    }


    //---------------------------------    SPECTRAL ANALYSIS   -------------------------------------//



    bool extractAtAEigenDecompositionWithInverseGeneralizedSystem( Eigen::Matrix< float_t, Eigen::Dynamic, Eigen::Dynamic > & evectors , unsigned int nVectorsToFind = 1 )
    {
        float_t epsilon = 0.00000000001;
        Eigen::SparseMatrix< float_t > diagonalMatrix(_AtA.rows(),_AtA.rows());
        {
            std::vector< Eigen::Triplet< float_t > > triplets;
            for( unsigned int r = 0 ; r < _AtA.rows() ; ++r ) triplets.push_back( Eigen::Triplet< float_t >(r,r,epsilon) );
            diagonalMatrix.setFromTriplets( triplets.begin() , triplets.end() );
        }

        Eigen::SparseMatrix< float_t > identityMatrix(_AtA.rows(),_AtA.rows());
        {
            std::vector< Eigen::Triplet< float_t > > triplets;
            for( unsigned int r = 0 ; r < _AtA.rows() ; ++r ) triplets.push_back( Eigen::Triplet< float_t >(r,r,1.0) );
            identityMatrix.setFromTriplets( triplets.begin() , triplets.end() );
        }

        // Construct matrix operation object using the wrapper classes
        Eigen::SparseMatrix< float_t > M = _AtA;

        M = M + diagonalMatrix;
        SparseCholesky< float_t >  op( M );
        SparseGenMatProd< float_t >  Bop(identityMatrix);

        // Construct generalized eigen solver object, requesting the largest nVectorsToFind generalized eigenvalues
        SymGEigsSolver< float_t , LARGEST_MAGN, SparseGenMatProd< float_t >, SparseCholesky< float_t >, GEIGS_CHOLESKY>
                geigs(&Bop, &op, nVectorsToFind, nVectorsToFind + 1);

        // Initialize and compute
        geigs.init();
        int nconv = geigs.compute();
        // Retrieve results
        if(geigs.info() == SUCCESSFUL)
        {
            evectors = geigs.eigenvectors();
            return true;
        }

        std::cout << "Eigenvalues WERE NOT found" << std::endl;
        return false;
    }


    bool extractAtAEigenDecompositionWithShiftInvert( Eigen::Matrix< float_t, Eigen::Dynamic, Eigen::Dynamic > & evectors , unsigned int nVectorsToFind = 1 )
    {
        // Construct matrix operation object using the wrapper class SparseGenMatProd

        Eigen::SparseMatrix< float_t > M = _AtA;
        float_t epsilon = 0.00000000001;
        Eigen::SparseMatrix< float_t > diagonalMatrix(_AtA.rows(),_AtA.rows());
        {
            std::vector< Eigen::Triplet< float_t > > triplets;
            for( unsigned int r = 0 ; r < _AtA.rows() ; ++r ) {
                triplets.push_back( Eigen::Triplet< float_t >(r,r,epsilon) );
            }
            diagonalMatrix.setFromTriplets( triplets.begin() , triplets.end() );
        }
        M = M + diagonalMatrix ;

        SparseSymShiftSolve< float_t > op( M );

        // Construct eigen solver object with shift 0
        // This will find eigenvalues that are closest to 0
        SymEigsShiftSolver< float_t, LARGEST_MAGN,
                SparseSymShiftSolve<float_t> > eigs(&op, nVectorsToFind, 20, 0.0);

        // Initialize and compute
        eigs.init();
        int nconv = eigs.compute();

        // Retrieve results
        if(eigs.info() == SUCCESSFUL) {
            evectors = eigs.eigenvectors();
            return true;
        }

        std::cout << "Eigenvalues WERE NOT found" << std::endl;
        return false;
    }
};

#endif // linearSystemSpectral_H
